package ru.dbotthepony.mc.otm.datagen.models

import com.google.gson.JsonObject
import com.google.gson.JsonParser
import net.minecraft.resources.ResourceLocation
import net.minecraft.server.packs.PackType
import net.minecraft.world.level.block.Block
import net.minecraftforge.client.model.generators.ModelProvider
import net.minecraftforge.data.event.GatherDataEvent
import ru.dbotthepony.mc.otm.datagen.DataGen
import ru.dbotthepony.mc.otm.core.registryName
import java.util.LinkedList

private typealias Callback = (MatteryModelProvider) -> Unit

sealed class MatteryModelProvider(event: GatherDataEvent, folder: String) : ModelProvider<MatteryModelBuilder>(event.generator, DataGen.MOD_ID, folder, ::MatteryModelBuilder, event.existingFileHelper) {
	private fun extendWithFolder(rl: ResourceLocation): ResourceLocation {
		return if (rl.path.contains("/")) rl else ResourceLocation(rl.namespace, folder + "/" + rl.path)
	}

	fun copy(destination: String, source: String): MatteryModelBuilder {
		val destinationLocation = extendWithFolder(if (destination.contains(":")) ResourceLocation(destination) else ResourceLocation(modid, destination))
		val sourceLocation = extendWithFolder(if (source.contains(":")) ResourceLocation("models/$source.json") else ResourceLocation(modid, "models/$source.json"))

		check(!generatedModels.containsKey(destinationLocation)) { "Model provider already contains model $destinationLocation" }
		existingFileHelper.trackGenerated(destinationLocation, MODEL)

		return factory.apply(destinationLocation).also {
			generatedModels[destinationLocation] = it
			val resource = existingFileHelper.getResource(sourceLocation, PackType.CLIENT_RESOURCES)
			val stream = resource.open()
			val reader = stream.reader()

			try {
				it.fromJson(JsonParser.parseReader(reader) as JsonObject)
			} finally {
				reader.close()
				stream.close()
			}
		}
	}

	private val callbacks = LinkedList<Callback>()

	fun exec(callback: Callback) {
		callbacks.add(callback)
	}

	override fun registerModels() {
		for (callback in callbacks) {
			callback(this)
		}
	}
}

class MatteryBlockModelProvider(event: GatherDataEvent) : MatteryModelProvider(event, BLOCK_FOLDER) {
	override fun getName(): String {
		return "Block Models: $modid"
	}

	fun decorativeGlassAll(blocks: Collection<Block>) {
		for (block in blocks) {
			exec {
				cubeAll(block.registryName!!.path, ResourceLocation(DataGen.MOD_ID, "block/decorative/${block.registryName!!.path}")).renderType("translucent")
			}
		}
	}

	fun decorativeGlassAll(vararg blocks: Block) {
		for (block in blocks) {
			exec {
				cubeAll(block.registryName!!.path, ResourceLocation(DataGen.MOD_ID, "block/decorative/${block.registryName!!.path}")).renderType("translucent")
			}
		}
	}

	fun decorativeCubeAll(vararg blocks: Block) {
		for (block in blocks) {
			exec {
				cubeAll(block.registryName!!.path, ResourceLocation(DataGen.MOD_ID, "block/decorative/${block.registryName!!.path}"))
			}
		}
	}

	fun decorativeCubeAll(subdir: String, vararg blocks: Block) {
		for (block in blocks) {
			exec {
				cubeAll(block.registryName!!.path, ResourceLocation(DataGen.MOD_ID, "block/decorative/${subdir}/${block.registryName!!.path}"))
			}
		}
	}

	fun decorativeCubeAll(subdir: String, suffix: String, vararg blocks: Block) {
		for (block in blocks) {
			exec {
				cubeAll(block.registryName!!.path, ResourceLocation(DataGen.MOD_ID, "block/decorative/${subdir}/${block.registryName!!.path}$suffix"))
			}
		}
	}

	fun decorativeCubeAll(blocks: Collection<Block>) {
		for (block in blocks) {
			exec {
				cubeAll(block.registryName!!.path, ResourceLocation(DataGen.MOD_ID, "block/decorative/${block.registryName!!.path}"))
			}
		}
	}

	fun decorativeCubeAll(block: Block, texture: String) {
		exec {
			cubeAll(block.registryName!!.path, ResourceLocation(DataGen.MOD_ID, "block/decorative/$texture"))
		}
	}

	fun column(block: Block, end: String, side: String) {
		exec {
			cubeColumn(block.registryName!!.path, ResourceLocation(DataGen.MOD_ID, end), ResourceLocation(DataGen.MOD_ID, side))
		}
	}

	fun decorativeColumn(it: Block, end: String, side: String) {
		column(it, "block/decorative/$end", "block/decorative/$side")
	}

	fun resourceCubeAll(vararg blocks: Block) {
		for (block in blocks) {
			exec {
				cubeAll(block.registryName!!.path, ResourceLocation(DataGen.MOD_ID, "block/resource/${block.registryName!!.path}"))
			}
		}
	}
}
