package ru.dbotthepony.mc.otm.android

import com.google.common.collect.ImmutableList
import it.unimi.dsi.fastutil.objects.Object2ObjectArrayMap
import net.minecraft.ChatFormatting
import net.minecraft.network.chat.Component
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.item.ItemStack
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.client.render.SkinElement
import ru.dbotthepony.mc.otm.container.iterator
import ru.dbotthepony.mc.otm.core.nonEmpty
import ru.dbotthepony.mc.otm.registry.MRegistry
import kotlin.collections.ArrayList

typealias ResearchCallback = ((research: AndroidResearch, feature: AndroidFeature) -> Unit)

@Suppress("unused")
class AndroidResearchBuilder(
	var experience: Int = 0,
) {
	private val items = ArrayList<ItemStack>()
	private val prerequisites = ArrayList<Pair<() -> ResourceLocation, Boolean>>()
	private val blockers = ArrayList<Pair<() -> ResourceLocation, Boolean>>()

	private val features = ArrayList<DeferredFeature>()
	private val resolvedFeatures = ArrayList<ResolvedFeature>()

	private class DeferredFeature(
		val id: ResourceLocation,
		val level: Int,
		val callbackResearched: ResearchCallback?
	)

	private class ResolvedFeature(
		val feature: AndroidFeatureType<*>,
		val level: Int,
		val callbackResearched: ResearchCallback?
	)

	var name: Component? = null
	var description: MutableList<Component>? = null
	var hasDescription = false
	var skinIcon: SkinElement? = null
	var iconText: Component? = null

	fun withIconText(icon: Component?): AndroidResearchBuilder {
		this.iconText = icon
		return this
	}

	fun withIconText() = withIconText(null)

	fun withIcon(icon: SkinElement): AndroidResearchBuilder {
		this.skinIcon = icon
		return this
	}

	fun withDescription(): AndroidResearchBuilder {
		this.hasDescription = true
		this.description = null
		return this
	}

	fun withExperience(experience: Int): AndroidResearchBuilder {
		this.experience = experience
		return this
	}

	fun withDescription(vararg description: Component): AndroidResearchBuilder {
		this.description = description.toMutableList()
		return this
	}

	fun withDescription(description: List<Component>): AndroidResearchBuilder {
		this.description = ArrayList<Component>(description.size).also { it.addAll(description) }
		return this
	}

	fun withName(name: Component): AndroidResearchBuilder {
		this.name = name
		return this
	}

	@JvmOverloads
	fun addPrerequisite(id: ResourceLocation, rigid: Boolean = false): AndroidResearchBuilder {
		prerequisites.add({ id } to rigid)
		return this
	}

	/**
	 * Please avoid having multiple prerequisites as case with more than 1 prerequisite does not have proper
	 * research tree render logic (yet).
	 */
	@JvmOverloads
	fun addPrerequisite(id: () -> ResourceLocation, rigid: Boolean = false): AndroidResearchBuilder {
		prerequisites.add(id to rigid)
		return this
	}

	@JvmOverloads
	fun addBlocker(id: ResourceLocation, rigid: Boolean = false): AndroidResearchBuilder {
		blockers.add({ id } to rigid)
		return this
	}

	@JvmOverloads
	fun addBlocker(id: () -> ResourceLocation, rigid: Boolean = false): AndroidResearchBuilder {
		blockers.add(id to rigid)
		return this
	}

	/**
	 * Please avoid having multiple prerequisites as case with more than 1 prerequisite does not have proper
	 * research tree render logic (yet).
	 */
	fun addPrerequisite(type: AndroidResearchType<*>) = addPrerequisite({ type.registryName ?: throw NullPointerException("Provided $type has no registryName defined") })
	fun addBlocker(type: AndroidResearchType<*>) = addBlocker({ type.registryName ?: throw NullPointerException("Provided $type has no registryName defined") })

	@JvmOverloads
	fun addFeatureResult(id: ResourceLocation, level: Int = 0, callback: ResearchCallback? = null): AndroidResearchBuilder {
		features.add(DeferredFeature(id, level, callback))
		return this
	}

	@JvmOverloads
	fun addFeatureResult(feature: AndroidFeatureType<*>, level: Int = 0, callback: ResearchCallback? = null): AndroidResearchBuilder {
		resolvedFeatures.add(ResolvedFeature(feature, level, callback))
		return this
	}

	fun addFeatureResult(id: ResourceLocation, callback: ResearchCallback): AndroidResearchBuilder {
		features.add(DeferredFeature(id, 0, callback))
		return this
	}

	fun addItem(cost: ItemStack): AndroidResearchBuilder {
		items.add(cost)
		return this
	}

	private class RememberResearchLevel(val level: Int?)

	fun build(): AndroidResearchType<AndroidResearch> {
		val items: List<ItemStack> = ImmutableList.builder<ItemStack>().let {
			for (item in this.items) {
				it.add(item.copy())
			}

			it.build()
		}

		val prerequisites: List<Pair<() -> ResourceLocation, Boolean>> = ImmutableList.copyOf(this.prerequisites)
		val blockers: List<Pair<() -> ResourceLocation, Boolean>> = ImmutableList.copyOf(this.blockers)
		val experience = this.experience

		val name = name?.copy()
		val description = description?.let {
			val builder = ImmutableList.builder<Component>()

			for (component in it) {
				builder.add(component.copy())
			}

			builder.build()
		}

		val hasDescription = hasDescription
		val defFeatures = ImmutableList.copyOf(features)
		val defResolvedFeatures = ImmutableList.copyOf(resolvedFeatures)

		val skinIcon = skinIcon
		val iconText = iconText?.copy()

		val features: List<ResolvedFeature> by lazy {
			val builder = ImmutableList.builder<ResolvedFeature>()

			for (df in defFeatures) {
				builder.add(ResolvedFeature(
					feature = MRegistry.ANDROID_FEATURES.getValue(df.id) ?: throw NoSuchElementException("Can't find android feature ${df.id}"),
					level = df.level,
					callbackResearched = df.callbackResearched
				))
			}

			for (df in defResolvedFeatures) {
				df.feature.registryName ?: throw NullPointerException("Feature ${df.feature} is missing registry name")
				builder.add(df)
			}

			builder.build()
		}

		return object : AndroidResearchType<AndroidResearch>(factory@{ it, capability ->
			return@factory object : AndroidResearch(it, capability) {
				override fun invalidateNetwork() {
					// NO-OP
				}

				override fun onUnResearch() {
					for (feature in features) {
						val level = oldResearchLevel[feature.feature]
						val get = capability.getFeature(feature.feature)

						if (level != null && get != null) {
							if (get.level == feature.level) {
								if (level.level == null) {
									capability.removeFeature(feature.feature)
								} else {
									get.level = level.level
								}
							}
						}
					}

					oldResearchLevel.clear()
				}

				private val oldResearchLevel = Object2ObjectArrayMap<AndroidFeatureType<*>, RememberResearchLevel>()

				override fun onResearched() {
					oldResearchLevel.clear()

					try {
						for (feature in features) {
							var get = capability.getFeature(feature.feature)

							if (get == null) {
								get = capability.addFeature(feature.feature)
								get.level = feature.level
								oldResearchLevel[feature.feature] = RememberResearchLevel(null)
							} else {
								if (get.level < feature.level) {
									oldResearchLevel[feature.feature] = RememberResearchLevel(feature.level)
									get.level = feature.level
								}
							}

							feature.callbackResearched?.invoke(this, get)
						}
					} catch(err: Throwable) {
						oldResearchLevel.clear()
						throw err
					}
				}

				override fun consumeResearchCost(simulate: Boolean): Boolean {
					if (capability.ply.abilities.instabuild) {
						// creative
						return true
					}

					if (!simulate && !consumeResearchCost(true)) {
						return false
					}

					if (experience > 0) {
						if (capability.ply.experienceLevel < experience) {
							return false
						}

						if (!simulate) {
							capability.ply.giveExperienceLevels(-experience)
						}
					}

					for (item in items) {
						var required = item.count
						val iterator = capability.ply.inventory.iterator().nonEmpty()

						for (invItem in iterator) {
							if (ItemStack.isSameItemSameTags(invItem, item)) {
								val toExtract = required.coerceAtMost(invItem.count)
								required -= toExtract

								if (!simulate) {
									if (toExtract == invItem.count) {
										iterator.remove()
									} else {
										invItem.count -= toExtract
									}
								}

								if (required <= 0) {
									break
								}
							}
						}

						if (required > 0) {
							return false
						}
					}

					return true
				}

				override fun refund(simulate: Boolean): Boolean {
					if (simulate) {
						return true
					}

					if (experience > 0) {
						capability.ply.giveExperienceLevels(experience)
					}

					for (item in items) {
						val leftover = item.copy()
						capability.ply.inventory.add(leftover)

						if (!leftover.isEmpty) {
							capability.ply.spawnAtLocation(leftover)
						}
					}

					return true
				}

				override val tooltipLines: List<Component> by lazy {
					val builder = ImmutableList.builder<Component>()

					builder.add(name ?: this.type.displayName)

					if (hasDescription) {
						builder.addAll(this.type.displayDescription)
					} else if (description != null) {
						builder.addAll(description)
					}

					builder.build()
				}

				override val screenTooltipHeader: Component
					get() = name ?: this.type.displayName

				override val screenTooltipLines: List<Component> get() {
					val builder = ArrayList<Component>()

					builder.add(name ?: this.type.displayName)

					if (hasDescription) {
						builder.addAll(this.type.displayDescription)
					} else if (description != null) {
						builder.addAll(description)
					}

					if (experience != 0) {
						builder.add(
							TranslatableComponent("otm.android_station.research.xp_cost", experience).withStyle(
							if (capability.ply.experienceLevel >= experience)
								ChatFormatting.DARK_GREEN
							else
								ChatFormatting.DARK_RED
						))
					}

					for (value in this.type.flatPrerequisites) {
						val instance = capability.getResearch(value)

						builder.add(
							TranslatableComponent("android_research.status.requires", instance.screenTooltipHeader).withStyle(
							if (instance.isResearched)
								ChatFormatting.DARK_GREEN
							else
								ChatFormatting.DARK_RED
						))
					}

					for (value in this.type.flatBlockedBy) {
						builder.add(TranslatableComponent("android_research.status.blocked_by", capability.getResearch(value).screenTooltipHeader).withStyle(ChatFormatting.DARK_RED))
					}

					for (value in this.type.flatBlocking) {
						builder.add(TranslatableComponent("android_research.status.blocks", capability.getResearch(value).screenTooltipHeader).withStyle(ChatFormatting.DARK_RED))
					}

					return builder
				}

				override val skinIcon: SkinElement? = skinIcon
				override val iconText: Component? = iconText
			}
		}) {
			override val displayName: Component
				get() = name ?: super.displayName

			override val displayDescription: List<Component>
				get() = description ?: super.displayDescription

			override val definedPrerequisites: List<AndroidResearchType<*>> by lazy {
				val builder = ImmutableList.builder<AndroidResearchType<*>>()

				for ((value, rigid) in prerequisites) {
					val get = MRegistry.ANDROID_RESEARCH.getValue(value())

					if (get == null && rigid) {
						throw NullPointerException("Unable to find research with name $value")
					} else if (get != null) {
						builder.add(get)
					}
				}

				builder.build()
			}

			override val definedBlockedBy: List<AndroidResearchType<*>> by lazy {
				val builder = ImmutableList.builder<AndroidResearchType<*>>()

				for ((value, rigid) in blockers) {
					val get = MRegistry.ANDROID_RESEARCH.getValue(value())

					if (get == null && rigid) {
						throw NullPointerException("Unable to find research with name $value")
					} else if (get != null) {
						builder.add(get)
					}
				}

				builder.build()
			}
		}
	}
}

