package ru.dbotthepony.mc.otm.recipe

import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import net.minecraft.core.NonNullList
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.Container
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.Items
import net.minecraft.world.item.crafting.Ingredient
import net.minecraft.world.item.crafting.Recipe
import net.minecraft.world.item.crafting.RecipeSerializer
import net.minecraft.world.item.crafting.RecipeType
import net.minecraft.world.level.Level
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.block.entity.PlatePressBlockEntity
import ru.dbotthepony.mc.otm.container.get
import ru.dbotthepony.mc.otm.registry.MRecipes
import ru.dbotthepony.mc.otm.core.registryName

class PlatePressRecipe(
	private val id: ResourceLocation,
	val input: Ingredient,
	val output: Ingredient,
	val count: Int,
	val workTime: Int = 200
) : Recipe<Container> {
	override fun matches(container: Container, p_44003_: Level): Boolean {
		if (output.isEmpty || input.isEmpty)
			return false

		return input.test(container[PlatePressBlockEntity.SLOT_INPUT])
	}

	private val outputStack: ItemStack by lazy {
		if (output.isEmpty || input.isEmpty) {
			ItemStack.EMPTY
		} else {
			val items = output.items
			val bestMatch = items.firstOrNull { it.item.registryName?.namespace == OverdriveThatMatters.MOD_ID } ?: items[0]
			bestMatch.copy().also { it.count = this.count }
		}
	}

	override fun getIngredients(): NonNullList<Ingredient> {
		if (input.isEmpty || output.isEmpty)
			return super.getIngredients()

		return NonNullList.of(Ingredient.EMPTY, input)
	}

	override fun isIncomplete(): Boolean {
		return !(input.items.any { it.item != Items.BARRIER } && output.items.any { it.item != Items.BARRIER })
	}

	override fun assemble(p_44001_: Container): ItemStack = outputStack.copy()
	override fun canCraftInDimensions(p_43999_: Int, p_44000_: Int) = true
	override fun getResultItem(): ItemStack = outputStack

	override fun getId() = id

	override fun getSerializer(): RecipeSerializer<*> {
		return PlatePressRecipeFactory
	}

	override fun getType(): RecipeType<PlatePressRecipe> = MRecipes.PLATE_PRESS
}

object PlatePressRecipeFactory : RecipeSerializer<PlatePressRecipe> {
	private val EMPTY = PlatePressRecipe(ResourceLocation(OverdriveThatMatters.MOD_ID, "empty"), Ingredient.EMPTY, Ingredient.EMPTY, 1)
	private val LOGGER = LogManager.getLogger()

	override fun fromJson(loc: ResourceLocation, obj: JsonObject): PlatePressRecipe {
		val input = try {
			Ingredient.fromJson(obj["input"] ?: throw IllegalStateException("Recipe $loc has no input field defined"))
		} catch (err: Throwable) {
			if (err.message?.lowercase()?.contains("unknown item tag") == true) {
				LOGGER.warn("Ignoring recipe Input of $loc deserialization error, defaulting to empty recipe")
				LOGGER.warn(err)
				return EMPTY
			} else {
				throw IllegalStateException("Input of $loc is malformed", err)
			}
		}

		val result = try {
			Ingredient.fromJson(obj["result"] ?: throw IllegalStateException("Recipe $loc has no result field defined"))
		} catch (err: Throwable) {
			if (err.message?.lowercase()?.contains("unknown item tag") == true) {
				LOGGER.warn("Ignoring recipe Output of $loc deserialization error, defaulting to empty recipe")
				LOGGER.warn(err)
				return EMPTY
			} else {
				throw IllegalStateException("Result of $loc is malformed", err)
			}
		}

		val workTime = (obj["work_time"] as? JsonPrimitive)?.let { return@let try {it.asInt} catch(err: Throwable) {throw IllegalStateException("Invalid work_time")} } ?: 200

		check(workTime >= 0) { "work_time of $loc does not make any sense" }

		val count = ((obj["result"] as JsonObject)["count"]  as? JsonPrimitive)?.let { return@let try {it.asInt} catch(err: Throwable) {throw IllegalStateException("Invalid result.count")} } ?: 1

		return PlatePressRecipe(loc, input, result, count, workTime)
	}

	override fun fromNetwork(loc: ResourceLocation, buff: FriendlyByteBuf): PlatePressRecipe {
		return PlatePressRecipe(loc, Ingredient.fromNetwork(buff), Ingredient.fromNetwork(buff), buff.readInt())
	}

	override fun toNetwork(buff: FriendlyByteBuf, recipe: PlatePressRecipe) {
		recipe.input.toNetwork(buff)
		recipe.output.toNetwork(buff)
		buff.writeInt(recipe.workTime)
	}
}
